@@ -90,7 +90,7 @@ pub struct Disabled;
 
 pub struct Unknown;
 
-// /// Generic $GpioX pin
+/// Generic $GpioX pin
 pub struct GpioPin<MODE> {
     pin: i32,
     _mode: PhantomData<MODE>,
@@ -106,6 +106,53 @@ where
             _mode: PhantomData,
         }
     }
+
+    fn get_input_level(&self) -> bool {
+        (unsafe { gpio_get_level(self.pin) } != 0)
+    }
+
+    #[cfg(not(feature = "ulp"))]
+    fn get_output_level(&self) -> bool {
+        let pin = self.pin as u32;
+
+        #[cfg(esp32c3)]
+        let is_set_high = unsafe { (*(GPIO_OUT_REG as *const u32) >> pin) & 0x01 != 0 };
+        #[cfg(not(esp32c3))]
+        let is_set_high = if pin <= 31 {
+            // GPIO0 - GPIO31
+            unsafe { (*(GPIO_OUT_REG as *const u32) >> pin) & 0x01 != 0 }
+        } else {
+            // GPIO32+
+            unsafe { (*(GPIO_OUT1_REG as *const u32) >> (pin - 32)) & 0x01 != 0 }
+        };
+
+        is_set_high
+    }
+
+    #[cfg(feature = "ulp")]
+    fn get_output_level(&self) -> bool {
+        (unsafe { gpio_get_output_level(self.pin) } != 0)
+    }
+
+    fn set_output_level(&mut self, on: bool) -> Result<(), EspError> {
+        esp_result!(unsafe { gpio_set_level(self.pin(), (on as u8).into()) }, ())
+    }
+
+    #[cfg(not(feature = "ulp"))]
+    pub fn get_drive_strength(&self) -> Result<DriveStrength, EspError> {
+        let mut cap: gpio_drive_cap_t = 0;
+
+        esp!(unsafe { gpio_get_drive_capability(self.pin(), &mut cap as *mut _) })?;
+
+        Ok(cap.into())
+    }
+
+    #[cfg(not(feature = "ulp"))]
+    pub fn set_drive_strength(&mut self, strength: DriveStrength) -> Result<(), EspError> {
+        esp!(unsafe { gpio_set_drive_capability(self.pin(), strength.into()) })?;
+
+        Ok(())
+    }
 }
 
 impl<MODE> Pin for GpioPin<MODE>
@@ -133,11 +180,84 @@ where
 
 impl InputPin for GpioPin<Input> {}
 
+impl embedded_hal::digital::v2::InputPin for GpioPin<Input> {
+    type Error = EspError;
+
+    fn is_high(&self) -> Result<bool, Self::Error> {
+        Ok(self.get_input_level())
+    }
+
+    fn is_low(&self) -> Result<bool, Self::Error> {
+        Ok(!self.get_input_level())
+    }
+}
+
 impl OutputPin for GpioPin<Output> {}
 
-impl OutputPin for GpioPin<InputOutput> {}
+impl embedded_hal::digital::v2::OutputPin for GpioPin<Output> {
+    type Error = EspError;
+
+    fn set_high(&mut self) -> Result<(), Self::Error> {
+        self.set_output_level(true)
+    }
+
+    fn set_low(&mut self) -> Result<(), Self::Error> {
+        self.set_output_level(false)
+    }
+}
+
+impl embedded_hal::digital::v2::StatefulOutputPin for GpioPin<Output> {
+    fn is_set_high(&self) -> Result<bool, Self::Error> {
+        Ok(self.get_output_level())
+    }
+
+    fn is_set_low(&self) -> Result<bool, Self::Error> {
+        Ok(!self.get_output_level())
+    }
+}
+
+impl embedded_hal::digital::v2::toggleable::Default for GpioPin<Output> {}
+
 impl InputPin for GpioPin<InputOutput> {}
 
+impl embedded_hal::digital::v2::InputPin for GpioPin<InputOutput> {
+    type Error = EspError;
+
+    fn is_high(&self) -> Result<bool, Self::Error> {
+        Ok(self.get_input_level())
+    }
+
+    fn is_low(&self) -> Result<bool, Self::Error> {
+        Ok(!self.get_input_level())
+    }
+}
+
+impl OutputPin for GpioPin<InputOutput> {}
+
+impl embedded_hal::digital::v2::OutputPin for GpioPin<InputOutput> {
+    type Error = EspError;
+
+    fn set_high(&mut self) -> Result<(), Self::Error> {
+        self.set_output_level(true)
+    }
+
+    fn set_low(&mut self) -> Result<(), Self::Error> {
+        self.set_output_level(false)
+    }
+}
+
+impl embedded_hal::digital::v2::StatefulOutputPin for GpioPin<InputOutput> {
+    fn is_set_high(&self) -> Result<bool, Self::Error> {
+        Ok(self.get_output_level())
+    }
+
+    fn is_set_low(&self) -> Result<bool, Self::Error> {
+        Ok(!self.get_output_level())
+    }
+}
+
+impl embedded_hal::digital::v2::toggleable::Default for GpioPin<InputOutput> {}
+
 /// Interrupt events
 ///
 /// *Note: ESP32 has a bug (3.14), which prevents correct triggering of interrupts when
@@ -198,6 +318,7 @@ pub struct PushPull;
 pub struct Analog;
 
 /// Drive strength (values are approximates)
+#[cfg(not(feature = "ulp"))]
 pub enum DriveStrength {
     I5mA = 0,
     I10mA = 1,
@@ -205,6 +326,32 @@ pub enum DriveStrength {
     I40mA = 3,
 }
 
+#[cfg(not(feature = "ulp"))]
+impl From<DriveStrength> for gpio_drive_cap_t {
+    fn from(strength: DriveStrength) -> gpio_drive_cap_t {
+        match strength {
+            DriveStrength::I5mA => gpio_drive_cap_t_GPIO_DRIVE_CAP_0,
+            DriveStrength::I10mA => gpio_drive_cap_t_GPIO_DRIVE_CAP_1,
+            DriveStrength::I20mA => gpio_drive_cap_t_GPIO_DRIVE_CAP_2,
+            DriveStrength::I40mA => gpio_drive_cap_t_GPIO_DRIVE_CAP_3,
+        }
+    }
+}
+
+#[cfg(not(feature = "ulp"))]
+impl From<gpio_drive_cap_t> for DriveStrength {
+    #[allow(non_upper_case_globals)]
+    fn from(cap: gpio_drive_cap_t) -> DriveStrength {
+        match cap {
+            gpio_drive_cap_t_GPIO_DRIVE_CAP_0 => DriveStrength::I5mA,
+            gpio_drive_cap_t_GPIO_DRIVE_CAP_1 => DriveStrength::I10mA,
+            gpio_drive_cap_t_GPIO_DRIVE_CAP_2 => DriveStrength::I20mA,
+            gpio_drive_cap_t_GPIO_DRIVE_CAP_3 => DriveStrength::I40mA,
+            other => panic!("Unknown GPIO pin drive capability: {}", other),
+        }
+    }
+}
+
 macro_rules! impl_hal_input_pin {
     ($pxi:ident: $mode:ident) => {
         impl embedded_hal::digital::v2::InputPin for $pxi<$mode> {
@@ -223,6 +370,24 @@ macro_rules! impl_hal_input_pin {
 
 macro_rules! impl_hal_output_pin {
     ($pxi:ident: $mode:ident) => {
+        impl $pxi<$mode> {
+            #[cfg(not(feature = "ulp"))]
+            pub fn get_drive_strength(&self) -> Result<DriveStrength, EspError> {
+                let mut cap: gpio_drive_cap_t = 0;
+
+                esp!(unsafe { gpio_get_drive_capability(self.pin(), &mut cap as *mut _) })?;
+
+                Ok(cap.into())
+            }
+
+            #[cfg(not(feature = "ulp"))]
+            pub fn set_drive_strength(&mut self, strength: DriveStrength) -> Result<(), EspError> {
+                esp!(unsafe { gpio_set_drive_capability(self.pin(), strength.into()) })?;
+
+                Ok(())
+            }
+        }
+
         impl embedded_hal::digital::v2::OutputPin for $pxi<$mode> {
             type Error = EspError;
 
